@@ -3,10 +3,14 @@ from struct import unpack
 from socket import inet_ntoa
 from ssl import CERT_NONE
 
+from Cryptodome.Cipher import AES
+from Cryptodome.Hash import MD4, SHA1
+from Cryptodome.Protocol.KDF import PBKDF2
+
 from ldap3 import Server, Connection, SASL, KERBEROS, NTLM, SUBTREE, ALL as LDAP3_ALL, BASE, DEREF_NEVER
 from ldap3 import SIMPLE
 from ldap3.protocol.formatters.formatters import format_sid
-from ldap3.core.exceptions import LDAPOperationResult, LDAPSocketOpenError, LDAPAttributeError
+from ldap3.core.exceptions import LDAPOperationResult, LDAPSocketOpenError, LDAPAttributeError, LDAPSocketSendError
 from ldap3.extend.microsoft.unlockAccount import ad_unlock_account
 from ldap3.extend.microsoft.modifyPassword import ad_modify_password
 from ldap3.extend.microsoft.addMembersToGroups import ad_add_members_to_groups as addUsersInGroups
@@ -15,10 +19,10 @@ from ldap3.extend.microsoft.removeMembersFromGroups import ad_remove_members_fro
 import ldap3
 
 from ldeep.views.activedirectory import ActiveDirectoryView, ALL, validate_sid, validate_guid
-from ldeep.views.constants import USER_ACCOUNT_CONTROL, DNS_TYPES, SAM_ACCOUNT_TYPE, PWD_PROPERTIES, TRUSTS_INFOS, WELL_KNOWN_SIDS, LOGON_SAM_LOGON_RESPONSE_EX
+from ldeep.views.constants import USER_ACCOUNT_CONTROL, DNS_TYPES, SAM_ACCOUNT_TYPE, PWD_PROPERTIES, TRUSTS_INFOS, WELL_KNOWN_SIDS, LOGON_SAM_LOGON_RESPONSE_EX, GMSA_ENCRYPTION_CONSTANTS
 from ldeep.utils.sddl import parse_ntSecurityDescriptor
-
-PAGE_SIZE = 1000
+from ldeep.utils.structure import Structure
+from ldeep.views.structures import MSDS_MANAGEDPASSWORD_BLOB 
 
 
 # define an ldap3-compliant formatters
@@ -68,6 +72,7 @@ def format_pwdProperties(raw_value):
         pass
     return raw_value
 
+
 # define an ldap3-compliant formatters
 def format_trustsInfos(raw_value):
     try:
@@ -99,13 +104,6 @@ def format_dnsrecord(raw_value):
                 target = ''.join([c for c in data if ord(c) > 31 or ord(c) == 9])
             return "%s %s" % (recordname, target)
 
-def format_ad_timedelta(raw_value):
-    """
-    Convert a negative filetime value to an integer timedelta.
-    """
-    if isinstance(raw_value, bytes):
-        raw_value = int(raw_value)
-    return raw_value
 
 def format_ad_timedelta(raw_value):
     """
@@ -116,7 +114,7 @@ def format_ad_timedelta(raw_value):
     return raw_value
 
 
-#from http://www.kouti.com/tables/baseattributes.htm
+# from http://www.kouti.com/tables/baseattributes.htm
 ldap3.protocol.formatters.standard.standard_formatter["1.2.840.113556.1.4.8"] = (format_userAccountControl, None)
 ldap3.protocol.formatters.standard.standard_formatter["1.2.840.113556.1.4.302"] = (format_samAccountType, None)
 ldap3.protocol.formatters.standard.standard_formatter["1.2.840.113556.1.4.382"] = (format_dnsrecord, None)
@@ -129,6 +127,7 @@ ldap3.protocol.formatters.standard.standard_formatter['1.2.840.113556.1.4.78'] =
 ldap3.protocol.formatters.standard.standard_formatter['1.2.840.113556.1.4.470'] = (format_trustsInfos, None)
 ldap3.protocol.formatters.standard.standard_formatter['1.2.840.113556.1.2.281'] = (parse_ntSecurityDescriptor, None)
 
+
 class LdapActiveDirectoryView(ActiveDirectoryView):
     """
     Manage a LDAP connection to a LDAP Active Directory.
@@ -141,7 +140,9 @@ class LdapActiveDirectoryView(ActiveDirectoryView):
     ZONE_FILTER = lambda _: "(objectClass=dnsNode)"
     SITES_FILTER = lambda _: "(objectClass=site)"
     SUBNET_FILTER = lambda _, s: f"(SiteObject={s})"
-    PKI_FILTER = lambda _: "(objectclass=pKIEnrollmentService)"
+    PKI_FILTER = lambda _: "(objectClass=pKIEnrollmentService)"
+    PRIMARY_SCCM_FILTER = lambda _: "(cn=System Management)"
+    DP_SCCM_FILTER = lambda _: "(objectClass=mssmsmanagementpoint)"
     USER_ALL_FILTER = lambda _: "(&(objectCategory=Person)(objectClass=user))"
     USER_SPN_FILTER = lambda _: "(&(objectCategory=Person)(objectClass=user)(servicePrincipalName=*)(!(sAMAccountName=krbtgt)))"
     USER_ACCOUNT_CONTROL_FILTER = lambda _, n: f"(&(objectCategory=Person)(objectClass=user)(userAccountControl:1.2.840.113556.1.4.803:={n}))"
@@ -167,7 +168,11 @@ class LdapActiveDirectoryView(ActiveDirectoryView):
     SILO_FILTER = lambda _, s: f"(&(objectClass=msDS-AuthNPolicySilo)(cn={s}))"
     LAPS_FILTER = lambda _, s: f"(&(objectCategory=computer)(ms-MCS-AdmPwd=*)(cn={s}))"
     SMSA_FILTER = lambda _: "(ObjectClass=msDS-ManagedServiceAccount)"
-    SHADOW_PRINCIPALS_FILTER = lambda _: "(objectclass=msDS-ShadowPrincipal)"
+    BITLOCKERKEY_FILTER = lambda _: "(objectClass=msFVE-RecoveryInformation)"
+    FSMO_DOMAIN_NAMING_FILTER = lambda _: "(&(objectClass=crossRefContainer)(fSMORoleOwner=*))"
+    FSMO_SCHEMA_FILTER = lambda _: "(&(objectClass=dMD)(fSMORoleOwner=*))"
+    FSMO_DOMAIN_FILTER = lambda _: "(fSMORoleOwner=*)"
+    SHADOW_PRINCIPALS_FILTER = lambda _: "(objectClass=msDS-ShadowPrincipal)"
     UNCONSTRAINED_DELEGATION_FILTER = lambda _: f"(userAccountControl:1.2.840.113556.1.4.803:=524288)"
     CONSTRAINED_DELEGATION_FILTER = lambda _: f"(msDS-AllowedToDelegateTo=*)"
     RESOURCE_BASED_CONSTRAINED_DELEGATION_FILTER = lambda _: f"(msDS-AllowedToActOnBehalfOfOtherIdentity=*)"
@@ -176,7 +181,7 @@ class LdapActiveDirectoryView(ActiveDirectoryView):
     class ActiveDirectoryLdapException(Exception):
         pass
 
-    def __init__(self, server, domain="", base="", username="", password="", ntlm="", pfx_file="", cert_pem="", key_pem="", method="NTLM"):
+    def __init__(self, server, domain="", base="", username="", password="", ntlm="", pfx_file="", pfx_pass="", cert_pem="", key_pem="", method="NTLM", throttle=0, page_size=1000):
         """
         LdapActiveDirectoryView constructor.
         Initialize the connection with the LDAP server.
@@ -200,11 +205,14 @@ class LdapActiveDirectoryView(ActiveDirectoryView):
         self.password = password
         self.ntlm = ntlm
         self.pfx_file = pfx_file
+        self.pfx_pass = pfx_pass
         self.cert = cert_pem
         self.key = key_pem
         self.server = server
         self.domain = domain
         self.hostnames = []
+        self.throttle = throttle
+        self.page_size = page_size
 
         self.set_controls()
         self.set_all_attributes()
@@ -216,18 +224,26 @@ class LdapActiveDirectoryView(ActiveDirectoryView):
                 exit(1)
             else:
                 if self.pfx_file:
-                    # TODO deal with pfx password protected
-                    with open(pfx_file, 'rb') as f:
-                        pfxdata = f.read()
                     from cryptography.hazmat.primitives.serialization import pkcs12
                     from cryptography.hazmat.primitives import serialization
-                    privkey, cert, extra_certs = pkcs12.load_key_and_certificates(pfxdata, None)
-                    key = privkey.private_bytes(
-                        encoding=serialization.Encoding.PEM,
-                        format=serialization.PrivateFormat.TraditionalOpenSSL,
-                        encryption_algorithm=serialization.NoEncryption(),
-                    )
-                    cert = cert.public_bytes(encoding=serialization.Encoding.PEM)
+                    with open(pfx_file, 'rb') as f:
+                        pfxdata = f.read()
+                    if self.pfx_pass:
+                        from oscrypto.keys import parse_pkcs12, parse_certificate, parse_private
+                        from oscrypto.asymmetric import rsa_pkcs1v15_sign, load_private_key, dump_openssl_private_key, dump_certificate
+                        if isinstance(self.pfx_pass, str):
+                            pfxpass = self.pfx_pass.encode()
+                        privkeyinfo, certinfo, _ = parse_pkcs12(pfxdata, password=pfxpass)
+                        key = dump_openssl_private_key(privkeyinfo, self.pfx_pass)
+                        cert = dump_certificate(certinfo, encoding='pem')
+                    else:
+                        privkey, cert, extra_certs = pkcs12.load_key_and_certificates(pfxdata, None)
+                        key = privkey.private_bytes(
+                            encoding=serialization.Encoding.PEM,
+                            format=serialization.PrivateFormat.TraditionalOpenSSL,
+                            encryption_algorithm=serialization.NoEncryption(),
+                        )
+                        cert = cert.public_bytes(encoding=serialization.Encoding.PEM)
                     try:
                         from tempfile import gettempdir
                         key_path = f'{gettempdir()}/ldeep_key'
@@ -235,14 +251,14 @@ class LdapActiveDirectoryView(ActiveDirectoryView):
                         with open(key_path, "wb") as f1, open(cert_path, "wb") as f2:
                             f1.write(key)
                             f2.write(cert)
-                        tls = ldap3.Tls(local_private_key_file=key_path, local_certificate_file=cert_path, validate=CERT_NONE)
                     except PermissionError:
                         print("Can't write key and cert to disk")
                         exit(1)
+                    tls = ldap3.Tls(local_private_key_file=key_path, local_certificate_file=cert_path, validate=CERT_NONE)
                 else:
                     tls = ldap3.Tls(local_private_key_file=self.key, local_certificate_file=self.cert, validate=CERT_NONE)
         else:
-            tls=ldap3.Tls(validate=CERT_NONE)
+            tls = ldap3.Tls(validate=CERT_NONE)
 
         if self.server.startswith("ldaps"):
             server = Server(
@@ -301,22 +317,34 @@ class LdapActiveDirectoryView(ActiveDirectoryView):
                     if self.pfx_file:
                         os.remove(key_path)
                         os.remove(cert_path)
-                except ldap3.core.exceptions.LDAPSocketOpenError:
+                except LDAPSocketOpenError:
                     print("Cannot get private key data, corrupted key or wrong passphrase ?")
                     if self.pfx_file:
                         os.remove(key_path)
                         os.remove(cert_path)
                     exit(1)
                 except Exception as e:
-                    print("Unhandled Exception")
+                    print(f"Unhandled Exception: {e}")
                     import traceback
                     traceback.print_exc()
                     exit(1)
             else:
                 if not self.ldap.bind():
                     raise self.ActiveDirectoryLdapException("Unable to bind with provided information")
+                if method == "anonymous":
+                    anon_base = self.ldap.request['base'].split(',')
+                    for i,item in enumerate(anon_base):
+                        if item.startswith("DC="):
+                            anon_base = ','.join(anon_base[i:])
+                            break
+                    self.ldap.search(search_base=anon_base, search_filter='(&(objectClass=domain))', search_scope='SUBTREE', attributes='*')
+
+                    if len(self.ldap.entries) == 0:
+                        raise self.ActiveDirectoryLdapException("Unable to retrieve information with anonymous bind")
         except LDAPSocketOpenError:
             raise self.ActiveDirectoryLdapException(f"Unable to open connection with {self.server}")
+        except LDAPSocketSendError:
+            raise self.ActiveDirectoryLdapException(f"Unable to open connection with {self.server}, maybe LDAPS is not enabled ?")
 
         self.base_dn = base or server.info.other["defaultNamingContext"][0]
         self.fqdn = ".".join(map(lambda x: x.replace("DC=", ''), self.base_dn.split(',')))
@@ -331,7 +359,8 @@ class LdapActiveDirectoryView(ActiveDirectoryView):
     def all_attributes(self):
         return self.attributes
 
-    def query(self, ldapfilter, attributes=[], base=None, scope=None):
+    # Not used anymore
+    def __query(self, ldapfilter, attributes=[], base=None, scope=None):
         """
         Perform a query to the LDAP server and return the results.
 
@@ -351,7 +380,7 @@ class LdapActiveDirectoryView(ActiveDirectoryView):
                 search_scope=scope or self.search_scope,
                 attributes=attributes,
                 controls=self.controls,
-                paged_size=PAGE_SIZE,
+                paged_size=self.page_size,
                 generator=True
             )
 
@@ -369,20 +398,57 @@ class LdapActiveDirectoryView(ActiveDirectoryView):
 
         return result_set
 
+    def query(self, ldapfilter, attributes=[], base=None, scope=None):
+        """
+        Perform a query to the LDAP server and return the results as a generator.
+
+        @ldapfilter: The LDAP filter to query (see RFC 2254).
+        @attributes: List of attributes to retrieved with the query.
+        @base: Base to use during the request.
+        @scope: Scope to use during the request.
+
+        @return a generator yielding records.
+        """
+        attributes = self.attributes if attributes == [] else attributes
+        # result_set = []
+        try:
+            entry_generator = self.ldap.extend.standard.paged_search(
+                search_base=base or self.base_dn,
+                search_filter=ldapfilter,
+                search_scope=scope or self.search_scope,
+                attributes=attributes,
+                controls=self.controls,
+                paged_size=self.page_size,
+                generator=True
+            )
+
+        except LDAPOperationResult as e:
+            raise self.ActiveDirectoryLdapException(e)
+        except LDAPAttributeError as e:
+            if not _getframe().f_back.f_code.co_name == "get_laps":
+                raise self.ActiveDirectoryLdapException(e)
+
+        def result(x):
+            if "dn" in x:
+                d = x["attributes"]
+                d["dn"] = x["dn"]
+                return dict(d)
+
+        return filter(lambda x: x is not None, map(result, entry_generator))
+
     def get_domain_sid(self):
         """
-        Return the current domain SID by issueing a LDAP request.
+        Return the current domain SID by issuing a LDAP request.
 
         @return the domain sid or None if a problem occurred.
         """
-        results = self.query(self.DOMAIN_INFO_FILTER(), ["ObjectSid"])
+        results = list(self.query(self.DOMAIN_INFO_FILTER(), ["ObjectSid"]))
 
         if results:
             return results[0]["objectSid"]
 
         return None
 
-
     def resolve_sid(self, sid):
         """
         Two cases:
@@ -453,6 +519,81 @@ class LdapActiveDirectoryView(ActiveDirectoryView):
 
         return result_set
 
+    def get_gmsa(self, attributes):
+
+        try:
+            self.ldap.start_tls()
+        except Exception as e:
+            print(f"Can't retrieve gmsa, TLS needed: {e}")
+            return []
+        entries = list(self.query("(ObjectClass=msDS-GroupManagedServiceAccount)", attributes))
+
+        constants = GMSA_ENCRYPTION_CONSTANTS
+        iv = b'\x00' * 16
+
+        for entry in entries:
+            sam = entry['sAMAccountName']
+            data = entry['msDS-ManagedPassword']
+            readers = entry['msDS-GroupMSAMembership']
+
+            # Find principals who can read the password
+            try:
+                readers_sd = parse_ntSecurityDescriptor(readers)
+                entry["readers"] = []
+                for ace in readers_sd["DACL"]["ACEs"]:
+                    try:
+                        reader_object = list(self.resolve_sid(ace["SID"]))
+                        if reader_object:
+                            name = reader_object[0]["sAMAccountName"]
+                            if "group" in reader_object[0]["objectClass"]:
+                                name += " (group)"
+                            entry["readers"].append(name)
+                        else:
+                            entry["readers"].append(ace["SID"])
+                    except Exception:
+                        pass
+            except Exception:
+                pass
+            
+            blob = MSDS_MANAGEDPASSWORD_BLOB()
+            try:
+                blob.fromString(data)
+            except (TypeError, KeyError):
+                continue
+
+            password = blob['CurrentPassword'][:-2]
+
+            # Compute NT hash
+            hash = MD4.new()
+            hash.update(password)
+            nthash = hash.hexdigest()
+
+            # Quick and dirty way to get the FQDN of the account's domain
+            dc_list = []
+            for s in entry['dn'].split(','):
+                if s.startswith('DC='):
+                    dc_list.append(s[3:])
+
+            domain_fqdn = '.'.join(dc_list)
+            salt = f'{domain_fqdn.upper()}host{sam[:-1].lower()}.{domain_fqdn.lower()}'
+            encryption_key = PBKDF2(password.decode('utf-16-le', 'replace').encode(), salt.encode(), 32, count=4096, hmac_hash_module=SHA1)
+
+            # Compute AES keys
+            cipher = AES.new(encryption_key, AES.MODE_CBC, iv)
+            first_part = cipher.encrypt(constants)
+            cipher = AES.new(encryption_key, AES.MODE_CBC, iv)
+            second_part = cipher.encrypt(first_part)
+            aes256_key = first_part[:16] + second_part[:16]
+
+            cipher = AES.new(encryption_key[:16], AES.MODE_CBC, iv)
+            aes128_key = cipher.encrypt(constants[:16])
+
+            entry["nthash"] = f"{nthash}"
+            entry["aes128-cts-hmac-sha1-96"] = f"{aes128_key.hex()}"
+            entry["aes256-cts-hmac-sha1-96"] = f"{aes256_key.hex()}"
+
+        return entries
+
     def unlock(self, username):
         """
         Unlock an account.
@@ -462,7 +603,7 @@ class LdapActiveDirectoryView(ActiveDirectoryView):
         @throw ActiveDirectoryLdapException if the account does not exist or the query returns more than one result.
         @return True if the account was successfully unlock or False otherwise.
         """
-        results = self.query(self.USER_DN_FILTER(username))
+        results = list(self.query(self.USER_DN_FILTER(username)))
         if len(results) != 1:
             raise self.ActiveDirectoryLdapException("Zero or non uniq result")
         else:
@@ -482,7 +623,7 @@ class LdapActiveDirectoryView(ActiveDirectoryView):
         @throw ActiveDirectoryLdapException if the account does not exist or the query returns more than one result.
         @return True if the account was successfully unlock or False otherwise.
         """
-        results = self.query(self.USER_DN_FILTER(username))
+        results = list(self.query(self.USER_DN_FILTER(username)))
         if len(results) != 1:
             raise self.ActiveDirectoryLdapException("Zero or non uniq result")
         else:
@@ -501,6 +642,7 @@ class LdapActiveDirectoryView(ActiveDirectoryView):
         try:
             return addUsersInGroups(self.ldap, user_dn, group_dn)
         except ldap3.core.exceptions.LDAPInvalidDnError as e:
+            print(f"Unhandled exception: {e}")
             # catch invalid group dn
             return False
 
@@ -516,6 +658,7 @@ class LdapActiveDirectoryView(ActiveDirectoryView):
         try:
             return removeUsersInGroups(self.ldap, user_dn, group_dn, fix=True)
         except ldap3.core.exceptions.LDAPInvalidDnError as e:
+            print(f"Unhandled exception: {e}")
             # catch invalid group dn
             return False
 
@@ -563,11 +706,11 @@ class LdapActiveDirectoryView(ActiveDirectoryView):
         computer_dn = f'CN={computer},CN=Computers,{self.base_dn}'
         # Default computer SPNs
         spns = [
-                    'HOST/%s' % computer,
-                    'HOST/%s.%s' % (computer, self.domain),
-                    'RestrictedKrbHost/%s' % computer,
-                    'RestrictedKrbHost/%s.%s' % (computer, self.domain),
-                ]
+            f"HOST/{computer}",
+            f"HOST/{computer}.{self.domain}",
+            f"RestrictedKrbHost/{computer}",
+            f"RestrictedKrbHost/{computer}.{self.domain}",
+        ]
 
         ucd = {
             'dnsHostName': '%s.%s' % (computer, self.domain),
@@ -577,7 +720,37 @@ class LdapActiveDirectoryView(ActiveDirectoryView):
             'unicodePwd': ('"%s"' % password).encode('utf-16-le')
         }
         try:
-            result = self.ldap.add(computer_dn, ['top','person','organizationalPerson','user','computer'], ucd)
+            result = self.ldap.add(computer_dn, ["top", "person", "organizationalPerson", "user", "computer"], ucd)
+        except Exception as e:
+            raise self.ActiveDirectoryLdapException(e)
+        return result
+
+    def create_user(self, user, password):
+        """
+        Create a user account on the domain.
+
+        @user: the name of the create user.
+        @password: the password of the user to create.
+
+        @return the result code on the add action
+        """
+        user_dn = f'CN={user},CN=Users,{self.base_dn}'
+
+        ucd = {
+            'objectCategory': 'CN=Person,CN=Schema,CN=Configuration,%s' % self.base_dn,
+            'distinguishedName': user_dn,
+            'cn': user,
+            'sn': user,
+            'givenName': user,
+            'displayName': user,
+            'name': user,
+            'userAccountControl': 0x200, # NORMAL_ACCOUNT (decimal value: 512)
+            'accountExpires': 0,
+            'sAMAccountName': user,
+            'unicodePwd': ('"%s"' % password).encode('utf-16-le')
+        }
+        try:
+            result = self.ldap.add(user_dn, ["top", "person", "organizationalPerson", "user"], ucd)
         except Exception as e:
             raise self.ActiveDirectoryLdapException(e)
         return result
